import platform

from dlt.common import logger
from dlt.destinations.exceptions import (
    DatabaseUndefinedRelation,
)
from dlt.destinations.impl.postgres.sql_client import Psycopg2SqlClient
from dlt.destinations.typing import DBTransaction

if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2  # type: ignore[import-not-found]
else:
    import psycopg2

from contextlib import contextmanager
from typing import Iterator

from dlt.destinations.sql_client import raise_database_error


class CrateDbSqlClient(Psycopg2SqlClient):
    """
    CrateDB SQL client, mostly compatible with PostgreSQL, with a few deviations.

    - Use `doc` as a default search path.
    - Disable transactions.
    """

    def open_connection(self) -> "psycopg2.connection":
        """
        Use `doc` instead of `public` as search path with CrateDB.
        """
        self._conn = psycopg2.connect(
            dsn=self.credentials.to_native_representation(),
            options=f"-c search_path={self.fully_qualified_dataset_name()},doc",
        )
        self._reset_connection()
        return self._conn

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        """
        CrateDB does not support transactions. Make emitting `BEGIN TRANSACTION` a no-op.
        """
        logger.warning(
            "CrateDB does not support transactions. Each SQL statement is auto-committed"
            " separately."
        )
        yield self

    @raise_database_error
    def commit_transaction(self) -> None:
        """
        CrateDB does not support transactions. Make emitting `COMMIT` a no-op.
        """
        pass

    @raise_database_error
    def rollback_transaction(self) -> None:
        """
        CrateDB does not support transactions. Raise an exception on `ROLLBACK`.

        TODO: Any better idea?
        """
        raise NotImplementedError("CrateDB statements can not be rolled back")

    @staticmethod
    def _is_error_schema_unknown(exception: Exception) -> bool:
        """
        CrateDB raises `Schema 'testdrive' unknown` errors when accessing schemas without tables.

        psycopg2.errors.InternalError_: Schema 'testdrive_staging' unknown
        dlt.destinations.exceptions.DatabaseTransientException: Schema 'testdrive_staging' unknown

        a) Try to ignore that.
        b) TODO: Refactor to synthesize an empty result, see `job_client_impl.get_stored_state`.
        c) Resolve in CrateDB.
        """
        msg = str(exception)
        return "Schema" in msg and "unknown" in msg

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        """
        Loop in additional error evaluation for CrateDB.
        """
        if cls._is_error_schema_unknown(exception=ex):
            raise DatabaseUndefinedRelation(ex)
        return Psycopg2SqlClient._make_database_exception(ex)

    def create_dataset(self) -> None:
        """
        CrateDB does not know `CREATE|DROP SCHEMA ...` statements.
        A no-op is not enough, because downstream operations expect the schema to exist.
        """
        self.execute_sql(
            f"CREATE TABLE IF NOT EXISTS {self.fully_qualified_dataset_name()}._placeholder (id"
            " INT)"
        )

    def drop_dataset(self) -> None:
        """
        CrateDB does not know `CREATE|DROP SCHEMA ...` statements.
        A no-op is not enough, because downstream operations expect the schema to exist.
        """
        self.execute_sql(f"DROP TABLE IF EXISTS {self.fully_qualified_dataset_name()}._placeholder")

    def _truncate_table_sql(self, qualified_table_name: str) -> str:
        """
        CrateDB does not understand `TRUNCATE TABLE ...`,
        and `caps.supports_truncate_command = False` is no longer honored?
        """
        return f"DELETE FROM {qualified_table_name} WHERE 1=1"  # noqa: S608



import json
from datetime import date, datetime, time
from typing import Any

from dlt.common.data_writers.escape import _escape_extended, _make_sql_escape_re

# CrateDB does not accept the original `{"'": "''", "\\": "\\\\", "\n": "\\n", "\r": "\\r"}`?
SQL_ESCAPE_DICT = {"'": "''"}
SQL_ESCAPE_RE = _make_sql_escape_re(SQL_ESCAPE_DICT)


def _escape_extended_cratedb(v: str) -> str:
    """
    The best-practice escaper for CrateDB, discovered by trial-and-error.
    """
    return _escape_extended(v, prefix="'", escape_dict=SQL_ESCAPE_DICT, escape_re=SQL_ESCAPE_RE)


def escape_cratedb_literal(v: Any) -> Any:
    """
    Based on `escape_postgres_literal`, with a mix of `escape_redshift_literal`.

    CrateDB needs a slightly adjusted escaping of literals.
    Examples: "L'Aupillon" and "Pizzas d'Anarosa" from `sys.summits`.

    It possibly also doesn't support the `E'` prefix as employed by the PostgreSQL escaper?
    Fortunately, the Redshift escaper came to the rescue, providing a reasonable baseline.

    CrateDB also needs support when serializing container types ARRAY vs. OBJECT.
    """
    if isinstance(v, str):
        return _escape_extended_cratedb(v)
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    # CrateDB, when serializing from an incoming `json` or `jsonb` type, the type mapper
    # can't know about what's actually inside, so arrays need a special treatment.
    if isinstance(v, list):
        v = {"array": v}
    if isinstance(v, dict):
        return _escape_extended_cratedb(json.dumps(v)) + "::OBJECT(DYNAMIC)"
    if isinstance(v, bytes):
        return f"'\\x{v.hex()}'"
    if v is None:
        return "NULL"

    return str(v)



from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Type

from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.schema.typing import TColumnSchema
from dlt.destinations.impl.postgres.factory import PostgresTypeMapper, postgres

from dlt_cratedb.impl.cratedb.configuration import CrateDbClientConfiguration
from dlt_cratedb.impl.cratedb.utils import escape_cratedb_literal

if TYPE_CHECKING:
    from dlt_cratedb.impl.cratedb.cratedb import CrateDbClient


class CrateDbTypeMapper(PostgresTypeMapper):
    """
    Adjust type mappings for CrateDB.

    - CrateDB uses `object(dynamic)` instead of `json` or `jsonb`.
    - CrateDB does not support `timestamp(6) without time zone`.
    - CrateDB does not support `time` for storing.
    - CrateDB does not support `binary` or `bytea`.
    """

    def __new__(cls, *args: List[Any], **kwargs: Dict[str, Any]) -> "CrateDbTypeMapper":
        cls.sct_to_unbound_dbt = deepcopy(PostgresTypeMapper.sct_to_unbound_dbt)
        cls.sct_to_unbound_dbt["json"] = "object(dynamic)"
        cls.sct_to_unbound_dbt["binary"] = "text"

        cls.sct_to_dbt = deepcopy(PostgresTypeMapper.sct_to_dbt)
        cls.sct_to_dbt["timestamp"] = "timestamp with time zone"
        del cls.sct_to_dbt["time"]

        cls.dbt_to_sct = deepcopy(PostgresTypeMapper.dbt_to_sct)
        cls.dbt_to_sct["jsonb"] = "object(dynamic)"  # type: ignore[assignment]
        cls.dbt_to_sct["bytea"] = "text"

        return super().__new__(cls)

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        """
        CrateDB does not support `timestamp(6) without time zone`.
        To not render the SQL clause like this, nullify the `precision` attribute.
        """
        column["precision"] = None
        return super().to_db_datetime_type(column, table)


class cratedb(postgres, Destination[CrateDbClientConfiguration, "CrateDbClient"]):
    spec = CrateDbClientConfiguration  # type: ignore[assignment]

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        """
        Tune down capabilities for CrateDB.
        """
        caps = super()._raw_capabilities()

        # CrateDB does not support transactions.
        caps.supports_transactions = False
        caps.supports_ddl_transactions = False

        # CrateDB does not support `TRUNCATE TABLE`, use `DELETE FROM` instead.
        caps.supports_truncate_command = False

        # CrateDB's type mapping needs adjustments compared to PostgreSQL.
        caps.type_mapper = CrateDbTypeMapper

        # TODO: Provide a dedicated dialect for SQLGlot.
        caps.sqlglot_dialect = "postgres"

        # CrateDB needs a slightly adjusted escaping of literals.
        # TODO: Escaping might need further adjustments, to be explored using integration tests.
        caps.escape_literal = escape_cratedb_literal

        # CrateDB does not support direct data loading using advanced formats.
        # TODO: Explore adding more formats for staged imports.
        caps.preferred_loader_file_format = "insert_values"
        caps.supported_loader_file_formats = ["insert_values"]
        caps.loader_file_format_selector = None

        return caps

    @property
    def client_class(self) -> Type["CrateDbClient"]:
        """
        Provide a different client for CrateDB.
        """
        from dlt_cratedb.impl.cratedb.cratedb import CrateDbClient

        return CrateDbClient


cratedb.register()



# ruff: noqa: S608
import logging
from typing import Any, Dict, List, Sequence

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    FollowupJobRequest,
    LoadJob,
    PreparedTableSchema,
)
from dlt.common.schema import Schema, TColumnHint
from dlt.destinations.impl.postgres.postgres import PostgresClient
from dlt.destinations.insert_job_client import InsertValuesJobClient
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.sql_jobs import SqlStagingReplaceFollowupJob

from dlt_cratedb.impl.cratedb.configuration import CrateDbClientConfiguration
from dlt_cratedb.impl.cratedb.sql_client import CrateDbSqlClient

# FIXME: The `UNIQUE` constraint is dearly missing.
#        When loading data multiple times, duplicates will happen.
HINT_TO_CRATEDB_ATTR: Dict[TColumnHint, str] = {"unique": ""}


logger = logging.getLogger(__name__)


class CrateDbStagingReplaceJob(SqlStagingReplaceFollowupJob):
    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        """
        CrateDB uses `ALTER CLUSTER SWAP TABLE`.

        -- https://github.com/crate/crate/issues/14833
        """
        sql: List[str] = []
        for table in table_chain:
            with sql_client.with_staging_dataset():
                staging_table_name = sql_client.make_qualified_table_name(table["name"])
            table_name = sql_client.make_qualified_table_name(table["name"])
            sql.extend(
                (
                    # Drop destination table.
                    f"DROP TABLE IF EXISTS {table_name};",
                    # Recreate destination table, because `ALTER CLUSTER SWAP TABLE` needs it.
                    (
                        f"CREATE TABLE IF NOT EXISTS {table_name} AS SELECT * FROM"
                        f" {staging_table_name} WHERE 1 = 0;"
                    ),
                    # Move the staging table to the destination schema.
                    f"ALTER CLUSTER SWAP TABLE {staging_table_name} TO {table_name};",
                    # CrateDB needs to flush writes.
                    f"REFRESH TABLE {table_name};",
                    # Recreate staging table not needed with CrateDB, because
                    # `ALTER CLUSTER SWAP TABLE` does not remove the source table.
                    (
                        f"CREATE TABLE IF NOT EXISTS {staging_table_name} AS SELECT * FROM"
                        f" {table_name} WHERE 1 = 0;"
                    ),
                    f"REFRESH TABLE {staging_table_name};",
                )
            )
        return sql


class CrateDbClient(PostgresClient):
    def __init__(
        self,
        schema: Schema,
        config: CrateDbClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        sql_client = CrateDbSqlClient(
            config.normalize_dataset_name(schema),
            config.normalize_staging_dataset_name(schema),
            config.credentials,
            capabilities,
        )
        InsertValuesJobClient.__init__(self, schema, config, sql_client)
        self.config: CrateDbClientConfiguration = config
        self.sql_client: CrateDbSqlClient = sql_client
        self.active_hints = HINT_TO_CRATEDB_ATTR if self.config.create_indexes else {}
        self.type_mapper = self.capabilities.get_type_mapper()

    def create_load_job(
        self,
        table: PreparedTableSchema,
        file_path: str,
        load_id: str,
        restore: bool = False,
    ) -> LoadJob:
        """
        CrateDB only supports the "insert values" paradigm.
        """
        job = InsertValuesJobClient.create_load_job(self, table, file_path, load_id, restore)
        if job is not None:
            return job
        return None

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        """
        CrateDB currently does not support "merge" followup jobs.
        -- https://github.com/crate/dlt-cratedb/issues/6

        Workaround: Redirect the "merge" job to use a "replace" job instead.
        """
        return [CrateDbStagingReplaceJob.from_table_chain(table_chain, self.sql_client)]

    def complete_load(self, load_id: str) -> None:
        """
        Intercept to invoke a `REFRESH TABLE ...` statement.
        """
        result = super().complete_load(load_id=load_id)
        table_name = self.sql_client.make_qualified_table_name(self.schema.loads_table_name)
        self.sql_client.execute_sql(f"REFRESH TABLE {table_name}")
        return result

    def _commit_schema_update(self, schema: Schema, schema_str: str) -> None:
        """
        Intercept to invoke a `REFRESH TABLE ...` statement.
        """
        result = super()._commit_schema_update(schema=schema, schema_str=schema_str)
        table_name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        self.sql_client.execute_sql(f"REFRESH TABLE {table_name}")
        return result

    def _delete_schema_in_storage(self, schema: Schema) -> None:
        """
        Intercept to invoke a `REFRESH TABLE ...` statement.
        """
        result = super()._delete_schema_in_storage(schema=schema)
        table_name = self.sql_client.make_qualified_table_name(self.schema.version_table_name)
        self.sql_client.execute_sql(f"REFRESH TABLE {table_name}")
        return result
